from lxml import html
from lxml import etree
import requests
import time
import urllib.request
import os
import datetime
import socket
from pathlib import Path
import re
import zipfile
from kol2unknownwords import pdf, collectionops
import json
import argparse
import unicodedata

def normalize(text):
    return unicodedata.normalize('NFC', text)

# Print iterations progress
def update_progressbar(iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█'):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end = '\r')
    # Print New Line on Complete
    if iteration == total:
        print()

def download_progress(blocknumber, blocksize, totalsize):
    if(totalsize > 0):
        update_progressbar(blocknumber * blocksize, totalsize)
    else:
        print('\r%.2f KBs' % (blocknumber * blocksize / 1024), end='\r')

def finish_progress():
    print('')

def normalize_filename(filename):
    return re.sub("\s\s+", " ", normalize(filename)).strip()

def httpget(url):
    while True:
        try:
            return requests.get(url)
        except socket.timeout:
            continue
        except requests.exceptions.ChunkedEncodingError:
            continue

def download(path, reprocess_all=False):
    print("Downloading TOPIK corpus.")
    socket.setdefaulttimeout(10)

    linkItemRegex = re.compile("btnView\('(?P<number>[0-9]+)'\)")
    documentLinkRegex = re.compile(
        "fileDownload2\('board.2','(?P<url>[^']+)','(?P<fileName>[^']+)'\); return false;")

    parser = html.HTMLParser(encoding="UTF-8")

    httpget("http://www.topik.go.kr/usr/cmm/subLocation.do?menuSeq=2210502")

    time.sleep(2)

    for pageIndex in range(1, 25):
        print("Downloading List Page {0}".format(pageIndex))
        formParams = {'boardSeq': '', 'confSeq': '2', 'cPage': str(pageIndex),
                      'pw': '', 'menu_seq': '2210502', 'searchKey': 'title',
                      'keyword': ''}
        while True:
            try:
                page = html.fromstring(
                    requests.post('http://www.topik.go.kr/usr/cmm/subLocation.do',
                                  formParams).content, parser=parser)
                break
            except socket.timeout:
                continue
        time.sleep(0.1)
        for folderLink in page.xpath("//td[@class='linkText']/a"):
            folderName = folderLink.text
            linkNumber = int(
                linkItemRegex.match(folderLink.xpath("@onclick")[0]).groups(
                    0)[0])
            print("Downloading Document Page {0}".format(linkNumber))
            formParams = {'boardSeq': str(linkNumber), 'confSeq': '2',
                          'cPage': str(pageIndex), 'pw': '',
                          'menu_seq': '2210502', 'searchKey': 'title',
                          'keyword': ''}
            while True:
                try:
                    documentPage = html.fromstring(
                        requests.post('http://www.topik.go.kr/usr/cmm/subLocation.do',
                                      formParams).content, parser=parser)
                    break
                except socket.timeout:
                    continue
                except requests.exceptions.ChunkedEncodingError:
                    continue
            time.sleep(0.1)
            for documentLink in documentPage.xpath(
                    "//div[@class='tblView01']//a/@onclick"):
                match = documentLinkRegex.match(documentLink)
                urlFileName = match.groups(0)[0]
                fileName = match.groups(0)[1].strip()
                targetFile = path / str(linkNumber) /\
                             normalize_filename(folderName) /\
                             normalize_filename(fileName)
                if fileName.endswith(".pdf") or fileName.endswith(".zip"):
                    placeholderFile = targetFile.parent / ("." + targetFile.name + ".deleted")
                    if (not targetFile.exists() and not placeholderFile.exists()) or reprocess_all:
                        targetFile.parent.mkdir(parents=True, exist_ok=True)
                        print("{0}".format(datetime.datetime.now()))
                        print("Downloading {0}".format(fileName))
                        done = False
                        while done == False:
                            try:
                                urllib.request.urlretrieve(
                                    "http://sco.topik.go.kr/topikdown.asp?file=" + urlFileName + "&path=upload_homepage/board.2",
                                    str(targetFile), download_progress)
                                finish_progress()
                                done = True
                            except socket.timeout:
                                continue
                            except requests.exceptions.ChunkedEncodingError:
                                continue;
                            finally:
                                if not done:
                                    os.remove(str(targetFile))
                    if targetFile.suffix == ".zip" and targetFile.exists():
                        with zipfile.ZipFile(str(targetFile), 'r') as archive:
                            for fileinfo in archive.infolist():
                                if fileinfo.filename.endswith(".pdf"):
                                    archive.extract(fileinfo, str(
                                        targetFile.parent / fileinfo.filename.replace(
                                            "/", "_").replace("\\", "_")))
                        with open(str(placeholderFile), "w") as file:
                            pass    #make empty placeholder file
                        targetFile.unlink()

text_extractor = pdf.TextExtractor()

def extract_text_from_pdf(path, download_path, txt_cahce_path, reprocess_all):
    txt_path = (txt_cahce_path / path).with_suffix(".txt")
    if not txt_path.exists() or reprocess_all:
        global text_extractor
        result = []
        for line in text_extractor.extract_from_file(normalize(str(download_path / path))):
            result.append(line)
        txt_path.parent.mkdir(parents=True, exist_ok=True)
        with open(str(txt_path), "w") as txt_file:
            txt_file.write("\n".join(result))
    with open(str(txt_path), "r") as txt_file:
        return txt_file.read()

def checksum(text):
    text = normalize(text)
    sum1 = sum2 = 0
    for byte in text.encode("utf-8"):
        sum1 = (sum1 + byte) % 255
        sum2 = (sum2 + sum1) % 255
    result = sum1 * 256 + sum2
    return result

def to_merged_xml(json_path, download_path, txt_cahce_path, target_path, reprocess_all):
    non_space_chars = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "~")
    with open(json_path, "r") as json_file:
        json_obj = json.load(json_file)
    annotators_element = etree.Element("annotators")
    for annotator in json_obj:
        annotator_element = etree.SubElement(annotators_element, "annotator")
        for key, value in annotator.items():
            if key != "passages":
                annotator_element.set(key, str(value))
        for passage in annotator["passages"]:
            passage_element = etree.SubElement(annotator_element, "passage")
            passage_element.set("source", passage["source"])
            passage_element.set("copyright", "National Institute of International Education (NIIED) in Korea")
            text = extract_text_from_pdf(passage["path"], download_path, txt_cahce_path, reprocess_all)
            stripped_text = re.sub(r"\s+", "", text)
            current_index = passage["offset"]
            passage_text = stripped_text[passage["offset"]:passage["offset"] + passage["length"]]
            if checksum(passage_text) != passage["checksum"]:
                raise RuntimeError(
                    "Matched text did not match the expected checksum.  This probably means the PDF's text could not be extracted correctly.")
            tokens = []
            for annotation in passage["annotations"]:
                start_index = annotation["offset"]
                if start_index > current_index:
                    unannotated_text = stripped_text[current_index : start_index]
                    if len(passage_element) > 0:
                        if unannotated_text[0] not in (",", "."):
                            unannotated_text = " " + unannotated_text
                        passage_element[-1].tail = unannotated_text
                    else:
                        passage_element.text = unannotated_text
                current_index = annotation["offset"] + annotation["length"]
                annotation_text = stripped_text[start_index : current_index]
                if len(passage_element) > 0:
                    if not passage_element[-1].tail:
                        passage_element[-1].tail = " "
                    elif passage_element[-1].tail[-1] not in non_space_chars:
                        passage_element[-1].tail += " "
                else:
                    if not passage_element.text:
                        passage_element.text = " "
                    elif passage_element.text[-1] not in non_space_chars:
                        passage_element.text += " "
                annotation_element = etree.SubElement(passage_element, "annotation")
                annotation_element.text = annotation_text
                annotation_element.set("unknown", str(annotation["unknown"]))
            if current_index < start_index + passage["length"]:
                passage_element[-1].tail = stripped_text[current_index:passage["offset"] + passage["length"]]
    with open(str(target_path), "wb") as target_file:
        target_file.write(etree.tostring(etree.ElementTree(annotators_element), pretty_print=True, encoding="utf-8"))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--reprocessall", action="store_true", help="Forces files to be downloaded and text to be extracted, even if the target file already exists.")
    parser.add_argument("--clean", action="store_true", help="Deletes the base directory before starting.")
    parser.add_argument("--input", "-i", default="./koreanL2unknownwords_annotatedcorpus_standoff.json", help="Path to the JSON file containing the standoff annotation data.")
    parser.add_argument("--topik_directory", "-t", default="./topik_ko", help="The path of the directory to download and store the TOPIK documents to.")
    parser.add_argument("--output", "-o", default="./koreanL2unknownwords_annotatedcorpus_merged.xml", help="The path of the output file to write the results to.")
    args = parser.parse_args()

    base_path = Path(args.topik_directory)
    download_path = base_path / "download"
    txt_cache_path = base_path / "txt"
    json_path = args.input
    target_path = args.output

    if args.clean:
        print("Deleting contents of {0} directory.".format(str(base_path)))
        def cleardir(path):
            for child in path.iterdir():
                if child.is_file():
                    child.unlink()
                else:
                    cleardir(child)
                    child.rmdir()
        cleardir(base_path)

    if input("This script will download past Test of Proficiency in Korean (TOPIK) "
          "exams from http://www.topik.go.kr/.  These texts are copyrighted by "
          "the National Institute for International Education (NIIED) in Korea.  To "
          "publish research that used this text, you must obtain permission "
          "from the copyright owner.  You may be able to contact and consult "
          "Seo ChaneWon at won2u@naver.com regarding the research approval. "
          "You are responsible for ensuring compliance with copyright laws "
          "regarding the TOPIK texts.  If have read and understood this message, "
          "please type \"Y\".\n").lower().startswith("y"):
        download_path.mkdir(parents=True, exist_ok=True)
        txt_cache_path.mkdir(parents=True, exist_ok=True)

        download(download_path, args.reprocessall)
        to_merged_xml(json_path, download_path, txt_cache_path, target_path, args.reprocessall)
    else:
        exit()
